import gleam/dynamic.{type Decoder}
import gleam/http/request.{type Request}
import gleam/http/response.{type Response}
import gleam/httpc
import gleam/json
import gleam/result
import gleam/string

import apkallone.{type Client, type Error}

pub fn prepare_request(
  client: Client,
  endpoint: String,
) -> Result(Request(String), Error) {
  use req <- result.try(
    request.to(client.instance <> endpoint)
    |> result.replace_error(apkallone.BadInstance),
  )

  req
  |> request.prepend_header("user-agent", apkallone.user_agent)
  |> Ok
}

pub fn send_request(req: Request(String)) -> Result(Response(String), Error) {
  use res <- result.try(
    httpc.send(req)
    |> result.map_error(apkallone.FetchError),
  )

  case res.status {
    200 -> Ok(res)
    other -> Error(apkallone.UnexpectedStatus(other))
  }
}

pub fn decode_response(
  res: Response(String),
  decoder: Decoder(a),
) -> Result(a, Error) {
  use <- require_content_type(res, "application/json")

  res.body
  |> json.decode(decoder)
  |> result.map_error(apkallone.DecodeError)
}

fn require_content_type(
  res: Response(String),
  content_type: String,
  fun: fn() -> Result(a, Error),
) -> Result(a, Error) {
  use actual_content_type <- result.try(
    response.get_header(res, "content-type")
    |> result.replace_error(apkallone.MissingContentType),
  )

  let actual_content_type =
    strip_parameters_from_content_type(actual_content_type)

  case actual_content_type == content_type {
    True -> fun()
    False -> Error(apkallone.UnexpectedContentType(actual_content_type))
  }
}

fn strip_parameters_from_content_type(content_type: String) -> String {
  case string.split_once(content_type, on: ";") {
    Ok(#(before_parameters, _parameters)) -> before_parameters
    Error(Nil) -> content_type
  }
}
